==> externalVector_1.4.0.tar.gz: externalVector/src/allocator.c <==
#include "common.h"

#define IF_ELSE_VECTOR_MODE(mode, block1, block2) \
    do { \
        switch (mode) { \
        case LGLSXP: \
        case INTSXP: \
        case REALSXP: \
        case CPLXSXP: \
        case STRSXP: \
        case VECSXP: \
        case CHARSXP: \
        { \
            block1; \
            break; \
        } \
        default: \
        { \
            block2; \
            break; \
        } \
        } \
    } while (0)

static SEXP
gcExternalPtr_new(SEXP size, SEXP type)
{
    SEXPTYPE mode = TYPEOF(type);
    int sz = asInteger(size);
    SEXP ans;

    IF_ELSE_VECTOR_MODE(mode, /* nothing */,
                        if (sz <= 0) {
                            return R_MakeExternalPtr(CHAR(R_BlankString),
                                                     R_NilValue,
                                                     R_BlankString);
                        }
                        mode = CHARSXP);
    ans = PROTECT(allocVector(mode, sz));
    /*
     * right now CHAR, LOGICAL, INTEGER, REAL ans COMPLEX all
     * return the same value
     */
    ans = R_MakeExternalPtr(CHAR(ans), R_NilValue, ans);
    UNPROTECT(1);
    return ans;
}

static SEXP
gcExternalPtr_size(SEXP ptr)
{
    void* addr = R_ExternalPtrAddr(ptr);
    SEXP store = R_ExternalPtrProtected(ptr);
    if (addr != R_NilValue && CHAR(store) == addr) {
        return ScalarInteger(LENGTH(store));
    }
    return ScalarInteger(NA_INTEGER);
}

static SEXP
gcExternalPtr_type(SEXP ptr)
{
    void* addr = R_ExternalPtrAddr(ptr);
    SEXP store = R_ExternalPtrProtected(ptr);
    if (addr == R_NilValue || CHAR(store) != addr) {
        error("memory not allocated by the gcAllocator");
    }
    return allocVector(TYPEOF(store), 1);
}

static SEXP
gcExternalPtr_restore(SEXP ptr)
{
    if (R_ExternalPtrAddr(ptr) == 0) {
        SEXP store = R_ExternalPtrProtected(ptr);
        IF_ELSE_VECTOR_MODE(TYPEOF(store),
                            R_SetExternalPtrAddr(ptr, CHAR(store)),
                            /* nothing */);
    }
    return ptr;
}

SEXP
gcExternalPtr_resizeInternal(SEXP ptr, int sz, int copy)
{
    void* addr = R_ExternalPtrAddr(ptr);
    SEXP oldstore = R_ExternalPtrProtected(ptr);


    if (addr != R_NilValue && CHAR(oldstore) == addr) {
        int oldsz = LENGTH(oldstore);

        if (sz < 0) sz = 0;

        if (sz != oldsz) {
            SEXPTYPE mode = TYPEOF(oldstore);

            if (mode == CHARSXP && sz == 0) {
                R_SetExternalPtrProtected(ptr, R_BlankString);
                R_SetExternalPtrAddr(ptr, CHAR(R_BlankString));
            } else {
                SEXP store = PROTECT(allocVector(mode, sz));
                sz = oldsz<sz?oldsz:sz;
                if (copy && sz > 0) {
                    int i;

                    switch (mode) {
                    case STRSXP:
                        for (i = 0; i < sz; i++)
                            SET_STRING_ELT(store, i,
                                           STRING_ELT(oldstore, i));
                        break;
                    case VECSXP:
                        for (i = 0; i < sz; i++)
                            SET_VECTOR_ELT(store, i,
                                           VECTOR_ELT(oldstore, i));
                        break;
                    default:
                        switch (mode) {
                        case LGLSXP:
                        case INTSXP:
                            sz *= sizeof(int);
                            break;
                        case REALSXP:
                            sz *= sizeof(double);
                            break;
                        case CPLXSXP:
                            sz *= sizeof(Rcomplex);
                            break;
                        case CHARSXP:
                            break;
                        default:
                            error("inconsistent allocated external pointer: report as a bug");
                            break;
                        }
                        memcpy(CHAR(store), CHAR(oldstore), sz);
                        break;
                    }
                }
                R_SetExternalPtrProtected(ptr, store);
                R_SetExternalPtrAddr(ptr, CHAR(store));
                UNPROTECT(1);
            }
        }
    }
    return ptr;
}

static SEXP
gcExternalPtr_resize(SEXP ptr, SEXP size, SEXP copy)
{
    return gcExternalPtr_resizeInternal(ptr, asInteger(size), asLogical(copy));
}

static SEXP
gcExternalPtr_delete(SEXP ptr)
{
    void* addr = R_ExternalPtrAddr(ptr);
    SEXP store = R_ExternalPtrProtected(ptr);
    if (addr != R_NilValue && CHAR(store) == addr) {
        R_SetExternalPtrAddr(ptr, R_NilValue);
        R_SetExternalPtrProtected(ptr, R_NilValue);
    }
    return R_NilValue;
}

#include <R.h>
#include <R_ext/Rdynload.h>

static const R_CallMethodDef R_CallDef  [] = {
    CALL_DEF(gcExternalPtr_new, 2),
    CALL_DEF(gcExternalPtr_type, 1),
    CALL_DEF(gcExternalPtr_restore, 1),
    CALL_DEF(gcExternalPtr_size, 1),
    CALL_DEF(gcExternalPtr_resize, 3),
    CALL_DEF(gcExternalPtr_delete, 1),
    {NULL, NULL, 0},
};

const R_CallMethodDef*
allocators_Call_methods(int* nptr)
{
    *nptr = LENGTH_OF_METHODS_ARRAY(R_CallDef);
    return &R_CallDef[0];
}

==> externalVector_1.6.0.tar.gz: externalVector/src/allocator.c <==
#include "common.h"

#define IF_ELSE_VECTOR_MODE(mode, block1, block2) \
    do { \
        switch (mode) { \
        case LGLSXP: \
        case INTSXP: \
        case REALSXP: \
        case CPLXSXP: \
        case STRSXP: \
        case VECSXP: \
        case CHARSXP: \
        { \
            block1; \
            break; \
        } \
        default: \
        { \
            block2; \
            break; \
        } \
        } \
    } while (0)

static SEXP
gcExternalPtr_new(SEXP size, SEXP type)
{
    SEXPTYPE mode = TYPEOF(type);
    int sz = asInteger(size);
    SEXP ans;

    IF_ELSE_VECTOR_MODE(mode, /* nothing */,
                        if (sz <= 0) {
                            return R_MakeExternalPtr(STRING_PTR(R_BlankString),
                                                     R_NilValue,
                                                     R_BlankString);
                        }
                        mode = CHARSXP);
    ans = PROTECT(allocVector(mode, sz));
    /*
     * right now CHAR, LOGICAL, INTEGER, REAL and COMPLEX all
     * return the same value
     */
    ans = R_MakeExternalPtr(STRING_PTR(ans), R_NilValue, ans);
    UNPROTECT(1);
    return ans;
}

static SEXP
gcExternalPtr_size(SEXP ptr)
{
    void* addr = R_ExternalPtrAddr(ptr);
    SEXP store = R_ExternalPtrProtected(ptr);
    if (addr != R_NilValue && CHAR(store) == addr) {
        return ScalarInteger(LENGTH(store));
    }
    return ScalarInteger(NA_INTEGER);
}

static SEXP
gcExternalPtr_type(SEXP ptr)
{
    void* addr = R_ExternalPtrAddr(ptr);
    SEXP store = R_ExternalPtrProtected(ptr);
    if (addr == R_NilValue || CHAR(store) != addr) {
        error("memory not allocated by the gcAllocator");
    }
    return allocVector(TYPEOF(store), 1);
}

static SEXP
gcExternalPtr_restore(SEXP ptr)
{
    if (R_ExternalPtrAddr(ptr) == 0) {
        SEXP store = R_ExternalPtrProtected(ptr);
        IF_ELSE_VECTOR_MODE(TYPEOF(store),
                            R_SetExternalPtrAddr(ptr, STRING_PTR(store)),
                            /* nothing */);
    }
    return ptr;
}

SEXP
gcExternalPtr_resizeInternal(SEXP ptr, int sz, int copy)
{
    void* addr = R_ExternalPtrAddr(ptr);
    SEXP oldstore = R_ExternalPtrProtected(ptr);


    if (addr != R_NilValue && CHAR(oldstore) == addr) {
        int oldsz = LENGTH(oldstore);

        if (sz < 0) sz = 0;

        if (sz != oldsz) {
            SEXPTYPE mode = TYPEOF(oldstore);

            if (mode == CHARSXP && sz == 0) {
                R_SetExternalPtrProtected(ptr, R_BlankString);
                R_SetExternalPtrAddr(ptr, STRING_PTR(R_BlankString));
            } else {
                SEXP store = PROTECT(allocVector(mode, sz));
                sz = oldsz<sz?oldsz:sz;
                if (copy && sz > 0) {
                    int i;

                    switch (mode) {
                    case STRSXP:
                        for (i = 0; i < sz; i++)
                            SET_STRING_ELT(store, i,
                                           STRING_ELT(oldstore, i));
                        break;
                    case VECSXP:
                        for (i = 0; i < sz; i++)
                            SET_VECTOR_ELT(store, i,
                                           VECTOR_ELT(oldstore, i));
                        break;
                    default:
                        switch (mode) {
                        case LGLSXP:
                        case INTSXP:
                            sz *= sizeof(int);
                            break;
                        case REALSXP:
                            sz *= sizeof(double);
                            break;
                        case CPLXSXP:
                            sz *= sizeof(Rcomplex);
                            break;
                        case CHARSXP:
                            break;
                        default:
                            error("inconsistent allocated external pointer: report as a bug");
                            break;
                        }
                        memcpy(STRING_PTR(store), STRING_PTR(oldstore), sz);
                        break;
                    }
                }
                R_SetExternalPtrProtected(ptr, store);
                R_SetExternalPtrAddr(ptr, STRING_PTR(store));
                UNPROTECT(1);
            }
        }
    }
    return ptr;
}

static SEXP
gcExternalPtr_resize(SEXP ptr, SEXP size, SEXP copy)
{
    return gcExternalPtr_resizeInternal(ptr, asInteger(size), asLogical(copy));
}

static SEXP
gcExternalPtr_delete(SEXP ptr)
{
    void* addr = R_ExternalPtrAddr(ptr);
    SEXP store = R_ExternalPtrProtected(ptr);
    if (addr != R_NilValue && CHAR(store) == addr) {
        R_SetExternalPtrAddr(ptr, R_NilValue);
        R_SetExternalPtrProtected(ptr, R_NilValue);
    }
    return R_NilValue;
}

#include <R.h>
#include <R_ext/Rdynload.h>

static const R_CallMethodDef R_CallDef  [] = {
    CALL_DEF(gcExternalPtr_new, 2),
    CALL_DEF(gcExternalPtr_type, 1),
    CALL_DEF(gcExternalPtr_restore, 1),
    CALL_DEF(gcExternalPtr_size, 1),
    CALL_DEF(gcExternalPtr_resize, 3),
    CALL_DEF(gcExternalPtr_delete, 1),
    {NULL, NULL, 0},
};

const R_CallMethodDef*
allocators_Call_methods(int* nptr)
{
    *nptr = LENGTH_OF_METHODS_ARRAY(R_CallDef);
    return &R_CallDef[0];
}
